from gwenflow.llms.base import ChatBase
from gwenflow.llms.azure import ChatAzureOpenAI
from gwenflow.llms.deepseek import ChatDeepSeek
from gwenflow.llms.google import ChatGemini
from gwenflow.llms.gwenlake import ChatGwenlake
from gwenflow.llms.mistral import ChatMistral
from gwenflow.llms.ollama import ChatOllama
from gwenflow.llms.openai import ChatOpenAI

__all__ = [
    "ChatBase",
    "ChatOpenAI",
    "ChatAzureOpenAI",
    "ChatGemini",
    "ChatMistral",
    "ChatGwenlake",
    "ChatOllama",
    "ChatDeepSeek",
]



import io
from pathlib import Path
from typing import List, Union

from gwenflow.logger import logger
from gwenflow.readers.base import Reader
from gwenflow.types import Document


class PDFReader(Reader):
    def read(self, file: Union[Path, io.BytesIO]) -> List[Document]:
        try:
            import pdfplumber
        except ImportError as e:
            raise ImportError("pdfplumber is not installed. Please install it with `pip install pdfplumber`.") from e

        try:
            filename = self.get_file_name(file)
            content = self.get_file_content(file)

            if isinstance(content, io.BytesIO):
                pdf_file = content
            else:
                pdf_file = io.BytesIO(content)

            documents = []
            with pdfplumber.open(pdf_file) as pdf:
                for i, page in enumerate(pdf.pages):
                    text = page.extract_text() or ""
                    safe_text = text.encode("utf-8", errors="ignore").decode("utf-8")

                    tables = page.extract_tables()

                    metadata = {
                        "filename": filename,
                        "page": i + 1,
                        "tables": tables,
                    }

                    doc = Document(id=self.key(f"{filename}_{i + 1}"), content=safe_text, metadata=metadata)
                    documents.append(doc)

        except Exception as e:
            logger.error(f"Error reading file: {e}")
            return []

        return documents



import json
from typing import Any

from openinference.semconv.trace import SpanAttributes
from opentelemetry.trace import Span


def prepare_llm_attributes(span: Span, instance: Any):
    if hasattr(instance, "model"):
        span.set_attribute(SpanAttributes.LLM_MODEL_NAME, str(instance.model))
    if hasattr(instance, "_model_params"):
        span.set_attribute(
            SpanAttributes.LLM_INVOCATION_PARAMETERS, json.dumps(instance._model_params, default=str)
        )

def capture_llm_usage(span: Span, result: Any):
    if not hasattr(result, "usage") or not result.usage:
        return

    usage = result.usage
    # TODO change later to add more details in the token comsuption
    if getattr(usage, "input_tokens", None) is not None:
        span.set_attribute(SpanAttributes.LLM_TOKEN_COUNT_PROMPT, usage.input_tokens)

    if getattr(usage, "output_tokens", None) is not None:
        span.set_attribute(SpanAttributes.LLM_TOKEN_COUNT_COMPLETION, usage.output_tokens)

    if getattr(usage, "total_tokens", None) is not None:
        span.set_attribute(SpanAttributes.LLM_TOKEN_COUNT_TOTAL, usage.total_tokens)

def capture_tool_calls(span: Span, tool_calls: Any):
    if not tool_calls:
        return
    try:
        tool_calls_dump = [tc.model_dump() for tc in tool_calls]
        tool_calls_json = json.dumps(tool_calls_dump, default=str)
        span.set_attribute("llm.tool_calls", tool_calls_json)

        return tool_calls_json
    except Exception:
        pass
    return None



import io
import json
import os
from collections import namedtuple
from operator import attrgetter

try:
    import boto3
except ImportError as error:
    raise ImportError("Please install boto3 with `pip install boto3`.") from error


def aws_get_instance_ip_address(instance_id, region_name="eu-west-3"):
    ec2 = boto3.resource("ec2", region_name=region_name)
    instance = ec2.Instance(instance_id)
    return instance.private_ip_address


def __prev_str(s):
    if len(s) == 0:
        return s
    s, c = s[:-1], ord(s[-1])
    if c > 0:
        s += chr(c - 1)
    s += "".join(["\u7fff" for _ in range(10)])
    return s


# Thanks to https://stackoverflow.com/questions/35803027/retrieving-subfolders-names-in-s3-bucket-from-boto3
s3object = namedtuple("S3Obj", ["key", "mtime", "size", "ETag"])


def aws_s3_list_files(bucket, path, start=None, end=None, recursive=True, list_dirs=True, list_objs=True, limit=None):
    """Iterator that lists a bucket's objects under path, (optionally) starting with start and ending before end.

    If recursive is False, then list only the "depth=0" items (dirs and objects).

    If recursive is True, then list recursively all objects (no dirs).

    Args:
        bucket:
            a boto3.resource('s3').Bucket().
        path:
            a directory in the bucket.
        start:
            optional: start key, inclusive (may be a relative path under path, or
            absolute in the bucket)
        end:
            optional: stop key, exclusive (may be a relative path under path, or
            absolute in the bucket)
        recursive:
            optional, default True. If True, lists only objects. If False, lists
            only depth 0 "directories" and objects.
        list_dirs:
            optional, default True. Has no effect in recursive listing. On
            non-recursive listing, if False, then directories are omitted.
        list_objs:
            optional, default True. If False, then directories are omitted.
        limit:
            optional. If specified, then lists at most this many items.

    Returns:
        an iterator of S3Obj.

    Examples:
        # set up
        >>> s3 = boto3.resource("s3")
        ... bucket = s3.Bucket("bucket-name")

        # iterate through all S3 objects under some dir
        >>> for p in aws_s3_list_files(bucket, "some/dir"):
        ...     print(p)

        # iterate through up to 20 S3 objects under some dir, starting with foo_0010
        >>> for p in aws_s3_list_files(bucket, "some/dir", limit=20, start="foo_0010"):
        ...     print(p)

        # non-recursive listing under some dir:
        >>> for p in aws_s3_list_files(bucket, "some/dir", recursive=False):
        ...     print(p)

        # non-recursive listing under some dir, listing only dirs:
        >>> for p in aws_s3_list_files(bucket, "some/dir", recursive=False, list_objs=False):
        ...     print(p)
    """
    kwargs = {}
    if start is not None:
        if not start.startswith(path):
            start = os.path.join(path, start)
        # note: need to use a string just smaller than start, because
        # the list_object API specifies that start is excluded (the first
        # result is *after* start).
        kwargs.update(Marker=__prev_str(start))
    if end is not None:
        if not end.startswith(path):
            end = os.path.join(path, end)
    if not recursive:
        kwargs.update(Delimiter="/")
        if not path.endswith("/") and len(path) > 0:
            path += "/"
    kwargs.update(Prefix=path)
    if limit is not None:
        kwargs.update(PaginationConfig={"MaxItems": limit})

    paginator = bucket.meta.client.get_paginator("list_objects")
    for resp in paginator.paginate(Bucket=bucket.name, **kwargs):
        q = []
        if "CommonPrefixes" in resp and list_dirs:
            q = [s3object(f["Prefix"], None, None, None) for f in resp["CommonPrefixes"]]
        if "Contents" in resp and list_objs:
            q += [s3object(f["Key"], f["LastModified"], f["Size"], f["ETag"]) for f in resp["Contents"]]
        # note: even with sorted lists, it is faster to sort(a+b)
        # than heapq.merge(a, b) at least up to 10K elements in each list
        q = sorted(q, key=attrgetter("key"))
        if limit is not None:
            q = q[:limit]
            limit -= len(q)
        for p in q:
            if end is not None and p.key >= end:
                return
            yield p


def aws_s3_read_file(bucket: str, file: str):
    client = boto3.client("s3")
    response = client.get_object(Bucket=bucket, Key=file)
    return response["Body"].read()


def aws_s3_delete_folder(bucket: str, prefix: str):
    if len(prefix) < 35:  # protect repo if prefix is not uuid4
        return False
    client = boto3.client("s3")
    for key in client.list_objects(Bucket=bucket, Prefix=prefix):
        key.delete()
    return True


def aws_s3_read_json_file(bucket: str, file: str):
    data = aws_s3_read_file(bucket=bucket, file=file)
    return json.loads(data.decode("utf-8"))


def aws_s3_read_text_file(bucket: str, file: str):
    data = aws_s3_read_file(bucket=bucket, file=file)
    return data.decode("utf-8")


def aws_s3_uri_to_bucket_key(uri: str):
    try:
        tmp = uri.replace("s3://", "")
        bucket = tmp.split("/")[0]
        key = tmp.replace(bucket, "").strip("/")
        return bucket, key
    except Exception:
        pass
    return None, None


def aws_s3_upload_fileobj(bucket: str, key: str, fileobj, content_type: str = "application/octet-stream") -> dict:
    client = boto3.client("s3")
    try:
        client.upload_fileobj(Fileobj=fileobj, Bucket=bucket, Key=key, ExtraArgs={"ContentType": content_type})
        return {"success": True, "bucket": bucket, "key": key, "content_type": content_type}
    except Exception as e:
        return {"success": False, "error": str(e)}


def aws_s3_delete_file(bucket: str, key: str) -> dict:
    client = boto3.client("s3")
    try:
        client.delete_object(Bucket=bucket, Key=key)
        return {"success": True, "bucket": bucket, "key": key}
    except Exception as e:
        return {"success": False, "error": str(e)}


def aws_s3_download_in_buffer(bucket: str, key: str):
    try:
        client = boto3.client("s3")
        buffer = io.BytesIO()
        client.download_fileobj(Bucket=bucket, Key=key, Fileobj=buffer)
        buffer.seek(0)
        return buffer
    except Exception as e:
        print(f"Error downloading s3://{bucket}/{key}: {e}")
        return None


def aws_s3_is_file(bucket: str, key: str) -> bool:
    client = boto3.client("s3")
    try:
        client.head_object(Bucket=bucket, Key=key)
        return True
    except Exception:
        return False



from typing import Any, Dict, List, Literal, Optional, Tuple, Union

from pydantic import BaseModel, ConfigDict, field_validator, model_validator

SYSTEM = "system"
USER = "user"
ASSISTANT = "assistant"
TOOL = "tool"


class ContentItem(BaseModel):
    text: Optional[str] = None
    image: Optional[str] = None
    file: Optional[str] = None
    audio: Optional[Union[str, dict]] = None
    video: Optional[Union[str, list]] = None

    @model_validator(mode="after")
    def check_exclusivity(self):
        provided_fields = 0
        if self.text is not None:
            provided_fields += 1
        if self.image:
            provided_fields += 1
        if self.file:
            provided_fields += 1
        if self.audio:
            provided_fields += 1
        if self.video:
            provided_fields += 1

        if provided_fields != 1:
            raise ValueError("Exactly one of 'text', 'image', 'file', 'audio', or 'video' must be provided.")
        return self

    def __repr__(self):
        return f"ContentItem({self.model_dump()})"

    def get_type_and_value(self) -> Tuple[Literal["text", "image", "file", "audio", "video"], str]:
        ((t, v),) = self.model_dump().items()
        assert t in ("text", "image", "file", "audio", "video")
        return t, v

    @property
    def type(self) -> Literal["text", "image", "file", "audio", "video"]:
        t, v = self.get_type_and_value()
        return t

    @property
    def value(self) -> str:
        t, v = self.get_type_and_value()
        return v


class Message(BaseModel):
    """Base message class."""

    role: str
    """The role of the messages author (system, user, assistant or tool)."""

    content: Optional[Union[str, List[Union[str, dict]], List[ContentItem]]] = None
    """Content of the message."""

    name: Optional[str] = None
    """An optional name."""

    tool_call_id: Optional[str] = None
    """Tool call that this message is responding to."""

    tool_calls: Optional[List[Dict[str, Any]]] = None
    """The tool calls generated by the model, such as function calls."""

    reasoning_content: Optional[str] = None
    """Thinking content from the model."""

    extra: Optional[dict] = None

    model_config = ConfigDict(extra="allow", populate_by_name=True, arbitrary_types_allowed=True)

    def __repr__(self):
        return f"Message({self.model_dump()})"

    @field_validator("role")
    @classmethod
    def role_checker(cls, value: str) -> str:
        if value not in [USER, ASSISTANT, SYSTEM, TOOL]:
            raise ValueError(f"{value} must be one of {','.join([USER, ASSISTANT, SYSTEM, TOOL])}")
        return value

    def to_dict(self, **kwargs: Any) -> Dict[str, Any]:
        message_dict = self.model_dump(**kwargs)
        message_dict = {
            k: v for k, v in message_dict.items() if v is not None and not (isinstance(v, (list, dict)) and len(v) == 0)
        }
        return message_dict

    def to_openai(self) -> Dict[str, Any]:
        """Format a message into the format expected by OpenAI."""
        message_dict: Dict[str, Any] = {
            "role": self.role,
            "content": self.content,
            "name": self.name,
            "tool_call_id": self.tool_call_id,
            "tool_calls": self.tool_calls,
        }
        message_dict = {k: v for k, v in message_dict.items() if v is not None}

        # OpenAI expects the tool_calls to be None if empty, not an empty list
        if self.tool_calls is not None and len(self.tool_calls) == 0:
            message_dict["tool_calls"] = None

        return message_dict

    # def to_function_call_dict(self) -> Dict[str, Any]:
    #     return {
    #         "content": self.content,
    #         "tool_call_id": self.tool_call_id,
    #         "tool_name": self.tool_name,
    #         "tool_args": self.tool_args,
    #         "tool_call_error": self.tool_call_error,
    #         "metrics": self.metrics,
    #         "created_at": self.created_at,
    #     }


from gwenflow.types.responses.base import Logprob, ReasoningItem, ResponseBase
from gwenflow.types.responses.response import (
    Response,
    ResponseContentItem,
    ResponseReasoningItem,
    ResponseToolCallItem,
)
from gwenflow.types.responses.response_event import (
    ResponseContentDeltaEvent,
    ResponseContentEvent,
    ResponseEvent,
    ResponseOutputItemEvent,
    ResponseReasoningDeltaEvent,
    ResponseReasoningEvent,
    ResponseToolCallEvent,
)

__all__ = [
    "Logprob",
    "ReasoningItem",
    "ResponseBase",
    "Response",
    "ResponseReasoningItem",
    "ResponseContentItem",
    "ResponseToolCallItem",
    "ResponseContentDeltaEvent",
    "ResponseContentEvent",
    "ResponseEvent",
    "ResponseOutputItemEvent",
    "ResponseReasoningDeltaEvent",
    "ResponseReasoningEvent",
    "ResponseToolCallEvent",
    ""
]



from abc import ABC, abstractmethod
from collections.abc import AsyncIterator
from typing import Any, Dict, Iterator, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from gwenflow.tools import BaseTool
from gwenflow.types import ModelResponse

LLM_MODEL_PARAMETERS = {
    # --- OPENAI ---
    "gpt-5.2": {
        "context_token": 400000,
        "reasoning": True
    },
    "gpt-5.2-pro": {
        "context_token": 400000,
        "reasoning": True
    },
    "gpt-5": {
        "context_token": 400000,
        "reasoning": True
    },
    "gpt-5-mini": {
        "context_token": 400000,
        "reasoning": True
    },
    "gpt-5-nano": {
        "context_token": 400000,
        "reasoning": True
    },
    "gpt-4.5": {
        "context_token": 128000,
        "reasoning": False
    },
    "gpt-4.1": {
        "context_token": 1047576,
        "reasoning": False
    },
    "o3-pro": {
        "context_token": 200000,
        "reasoning": True
    },
    "o3-mini-high": {
        "context_token": 128000,
        "reasoning": True
    },
    "o4-mini": {
        "context_token": 128000,
        "reasoning": True
    },
    "gpt-4o": {
        "context_token": 128000,
        "reasoning": False
    },
    "gpt-4o-mini": {
        "context_token": 128000,
        "reasoning": False
    },
    "o1-preview": {
        "context_token": 128000,
        "reasoning": True
    },
    "o1-mini": {
        "context_token": 128000,
        "reasoning": True
    },


    # --- DEEPSEEK ---
    "deepseek-chat": {
        "context_token": 128000,
        "reasoning": False
    },
    "deepseek-r1": {
        "context_token": 128000,
        "reasoning": True
    },

    # --- META ---
    "llama-3.1-70b-versatile": {
        "context_token": 131072,
        "reasoning": False
    },
    "llama-3.1-8b-instant": {
        "context_token": 131072,
        "reasoning": False
    },

    # --- MISTRAL ---
    "mixtral-8x7b-32768": {
        "context_token": 32768,
        "reasoning": False
    },
}


class ChatBase(BaseModel, ABC):
    model: str
    """The model to use when invoking the LLM."""

    system_prompt: Optional[str] = None
    """The system prompt to use when invoking the LLM."""

    tools: List[BaseTool] = Field(default_factory=list)
    """A list of tools that the LLM can use."""

    tool_choice: Optional[Union[str, Dict[str, Any]]] = None
    tool_type: str = Field(default="fncall")

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    @abstractmethod
    def invoke(self, *args, **kwargs) -> ModelResponse:
        pass

    @abstractmethod
    async def ainvoke(self, *args, **kwargs) -> ModelResponse:
        pass

    @abstractmethod
    def stream(self, *args, **kwargs) -> Iterator[ModelResponse]:
        pass

    @abstractmethod
    async def astream(self, *args, **kwargs) -> AsyncIterator[ModelResponse]:
        pass

    def get_context_window_size(self) -> int:
        # Only using 75% of the context window size to avoid cutting the message in the middle
        return int(LLM_MODEL_PARAMETERS.get(self.model, {}).get("context_token", 128000) * 0.75)

    def get_reasoning_model(self) -> str:
        return LLM_MODEL_PARAMETERS.get(self.model, {}).get("reasoning", False)

    def get_tool_names(self):
        return [tool.name for tool in self.tools]

    def get_tool_map(self):
        return {tool.name: tool for tool in self.tools}


from gwenflow.tools.base import BaseTool
from gwenflow.tools.duckduckgo import DuckDuckGoNewsTool, DuckDuckGoSearchTool
from gwenflow.tools.function import FunctionTool
from gwenflow.tools.mcp.tool import MCPTool
from gwenflow.tools.pdf import PDFReaderTool
from gwenflow.tools.retriever import RetrieverTool
from gwenflow.tools.tavily import TavilyWebSearchTool
from gwenflow.tools.website import WebsiteReaderTool
from gwenflow.tools.wikipedia import WikipediaTool
from gwenflow.tools.yahoofinance import (
    YahooFinanceNews,
    YahooFinanceScreen,
    YahooFinanceStock,
)

__all__ = [
    "BaseTool",
    "FunctionTool",
    "RetrieverTool",
    "WikipediaTool",
    "WebsiteReaderTool",
    "PDFReaderTool",
    "DuckDuckGoSearchTool",
    "DuckDuckGoNewsTool",
    "YahooFinanceNews",
    "YahooFinanceStock",
    "YahooFinanceScreen",
    "TavilyWebSearchTool",
    "MCPTool",
]



import functools
import json
from typing import Any

from mcp.types import Tool

from gwenflow.exceptions import GwenflowException, UserError
from gwenflow.logger import logger
from gwenflow.tools import BaseTool, FunctionTool
from gwenflow.tools.mcp.server import MCPServer


class MCPUtil:
    """Set of utilities for interop between MCP and Gwenflow tools."""

    @classmethod
    async def get_all_function_tools(cls, servers: list["MCPServer"]) -> list[BaseTool]:
        """Get all function tools from a list of MCP servers."""
        tools = []
        tool_names: set[str] = set()
        for server in servers:
            server_tools = await cls.get_function_tools(server)
            server_tool_names = {tool.name for tool in server_tools}
            if len(server_tool_names & tool_names) > 0:
                raise UserError(f"Duplicate tool names found across MCP servers: {server_tool_names & tool_names}")
            tool_names.update(server_tool_names)
            tools.extend(server_tools)

        return tools

    @classmethod
    async def get_function_tools(cls, server: "MCPServer") -> list[BaseTool]:
        """Get all function tools from a single MCP server."""
        tools = await server.list_tools()
        return [cls.to_function_tool(tool, server) for tool in tools]

    @classmethod
    def to_function_tool(cls, tool: "Tool", server: "MCPServer") -> FunctionTool:
        """Convert an MCP tool to a function tool."""
        return FunctionTool(
            name=tool.name,
            description=tool.description or "",
            params_json_schema=tool.inputSchema,
            func=functools.partial(cls.invoke_mcp_tool, server, tool),
            tool_type="function",
        )

    @classmethod
    async def invoke_mcp_tool(cls, server: "MCPServer", tool: "Tool", arguments: dict[str, Any] | None) -> str:
        """Invoke an MCP tool and return the result as a string."""
        try:
            result = await server.call_tool(tool.name, arguments)
        except Exception as e:
            logger.error(f"Error invoking MCP tool {tool.name}: {e}")
            raise GwenflowException(f"Error invoking MCP tool {tool.name}: {e}") from e

        logger.debug(f"MCP tool {tool.name} returned {result}")

        # The MCP tool result is a list of content items, whereas OpenAI tool outputs are a single
        # string. We'll try to convert.
        if len(result.content) == 1:
            tool_output = result.content[0].model_dump_json()
        elif len(result.content) > 1:
            tool_output = json.dumps([item.model_dump() for item in result.content])
        else:
            logger.error(f"Errored MCP tool result: {result}")
            tool_output = "Error running tool."

        return tool_output



from typing import Dict, List

import requests
from pydantic import model_validator

from gwenflow.api import api
from gwenflow.logger import logger
from gwenflow.reranker.base import Reranker
from gwenflow.types import Document


class GwenlakeReranker(Reranker):
    """Gwenlake reranker."""

    @model_validator(mode="before")
    @classmethod
    def validate_environment(cls, values: Dict) -> Dict:
        if "model" not in values:
            values["model"] = "BAAI/bge-reranker-v2-m3"
        return values

    def _rerank(self, query: str, input: List[str]) -> List[List[float]]:
        try:
            payload = {"query": query, "input": input, "model": self.model}
            response = api.client.post("/v1/rerank", json=payload)
        except requests.exceptions.RequestException as e:
            raise ValueError(f"Error raised by inference endpoint: {e}") from e

        if response.status_code != 200:
            raise ValueError(f"Error raised by inference API: rate limit exceeded.\nResponse: {response.text}")

        parsed_response = response.json()
        if "data" not in parsed_response:
            raise ValueError("Error raised by inference API.")

        reranking = []
        for e in parsed_response["data"]:
            reranking.append(e)

        return reranking

    def rerank(self, query: str, documents: List[Document]) -> List[Document]:
        if not documents:
            return []

        batch_size = 100
        reranked_documents = []
        try:
            for i in range(0, len(documents), batch_size):
                i_end = min(len(documents), i + batch_size)
                batch = documents[i:i_end]
                batch_processed = []
                for document in batch:
                    batch_processed.append(document.content)
                reranked_documents += self._rerank(query=query, input=batch_processed)
        except Exception as e:
            logger.error(e)
            return None

        if len(reranked_documents) > 0:
            compressed_documents = documents.copy()

            for i, _ in enumerate(compressed_documents):
                compressed_documents[i].score = reranked_documents[i]["relevance_score"]

            # Order by relevance score
            compressed_documents.sort(
                key=lambda x: x.score if x.score is not None else float("-inf"),
                reverse=True,
            )

            if self.top_k is not None:
                compressed_documents = compressed_documents[: self.top_k]

            if self.threshold is not None:
                compressed_documents = [d for d in compressed_documents if d.score > self.threshold]

            return compressed_documents

        return []



from typing import Any, List, Literal, Optional, Union

from pydantic import BaseModel, Field, field_validator

from gwenflow.tools import BaseTool
from gwenflow.types.responses import ReasoningItem
from gwenflow.types.usage import Usage


class ResponseReasoningItem(BaseModel):
    id: str
    type: Literal['reasoning']
    summary: Optional[str] = None

    @field_validator('summary', mode='before')
    @classmethod
    def extract_summary_text(cls, v: Any) -> Any:
        if v is None:
            return None
        item = v[0]
        return item.get("text")


class ResponseToolCallItem(BaseModel):
    id: str
    status: Optional[Literal['in_progress', 'searching', 'completed']] = None
    type: Optional[str] = None


class ResponseContentItem(BaseModel):
    id: str
    type: Literal['message']
    content: Optional[str] = None

    @field_validator('content', mode='before')
    @classmethod
    def extract_content_text(cls, v: Any) -> Optional[str]:
        if v is None:
            return None

        if isinstance(v, list) and len(v) > 0:
            first_item = v[0]

            if hasattr(first_item, 'text'):
                return first_item.text

            if isinstance(first_item, dict):
                return first_item.get('text')
        if isinstance(v, str):
            return v

        return None


class Response(BaseModel):
    id: str
    created_at: float
    completed_at: Optional[float] = None
    object: Literal["response"]
    model: str
    status: Literal['in_progress', 'completed']
    usage: Usage
    output: List[Union[ResponseReasoningItem, ResponseContentItem, ResponseToolCallItem]] = Field(default_factory=list)
    reasoning: Optional[ReasoningItem] = None
    text_format: Optional[Any] = None
    tools: Optional[List[Any]] = None
    tool_choice: Optional[Literal['none', 'auto', 'required']] = None
    temperature: Optional[float] = None
    top_p : Optional[float] = None

    def get_text(self) -> str:
        for item in self.output:
            if isinstance(item, ResponseContentItem) and item.content:
                return item.content
        return ""

    def get_reasoning(self) -> str:
        for item in self.output:
            if isinstance(item, ResponseReasoningItem) and item.summary:
                return item.summary
        return ""



import io
from pathlib import Path
from typing import ClassVar, List, Union

from gwenflow.logger import logger
from gwenflow.readers.base import Reader
from gwenflow.types import Document


class DocxReader(Reader):
    trans: ClassVar[dict[int, int | None]] = {
        0x00A0: 0x20,
        0x202F: 0x20,
        0x2007: 0x20,
        0x200B: None,
        0x200C: None,
        0x200D: None,
        0xFEFF: None,
    }

    def get_text(self, file_obj) -> str:
        try:
            import docx
        except ImportError as e:
            raise ImportError("python-docx is not installed. Please install it with `pip install python-docx`") from e
        doc = docx.Document(file_obj)
        return "\n".join((p.text.translate(self.trans) if p.text else "") for p in doc.paragraphs)

    def get_tables(self, file_obj):
        try:
            import docx
        except ImportError as e:
            raise ImportError("python-docx is not installed. Please install it with `pip install python-docx`") from e
        doc = docx.Document(file_obj)

        tables = []
        for t in doc.tables:
            rows = []
            for r in t.rows:
                cells = []
                for c in r.cells:
                    txt = "\n".join(p.text for p in c.paragraphs) if c.paragraphs else ""
                    txt = txt.translate(self.trans) if txt else ""
                    cells.append(txt)
                rows.append(cells)
            tables.append(rows)
        return tables

    def read(self, file: Union[Path, io.BytesIO]) -> List[Document]:
        try:
            filename = self.get_file_name(file)
            content = self.get_file_content(file)

            data = content.getvalue() if isinstance(content, io.BytesIO) else content

            text = self.get_text(io.BytesIO(data))
            tables = self.get_tables(io.BytesIO(data))

            doc = Document(
                id=self.key(f"{filename}"),
                content=text,
                metadata={"filename": filename, "tables": tables},
            )
            return [doc]

        except Exception as e:
            logger.exception(f"Error reading file: {e}")
            return []



from pathlib import Path
from typing import List

from gwenflow.logger import logger
from gwenflow.readers.base import Reader
from gwenflow.types import Document


class TextReader(Reader):
    def read(self, file: Path) -> List[Document]:
        try:
            filename = self.get_file_name(file)
            content = self.get_file_content(file, text_mode=True)

            documents = [
                Document(
                    id=self.key(filename),
                    content=content,
                    metadata={"filename": filename},
                )
            ]

        except Exception as e:
            logger.error(f"Error reading file: {e}")
            return []

        return documents



import asyncio
import json
from abc import ABC, abstractmethod
from typing import Any

from pydantic import BaseModel, model_validator

from gwenflow.tools.utils import function_to_json


class BaseTool(BaseModel, ABC):
    name: str
    """The unique name of the tool that clearly communicates its purpose."""

    description: str
    """Used to tell the model how to use the tool."""

    params_json_schema: dict[str, Any] = None
    """The JSON schema for the tool's parameters."""

    tool_type: str = "base"
    """Tool type: base, function, langchain."""

    max_results: int = 50
    """A max result for the tools."""

    @model_validator(mode="after")
    def model_valid(self) -> Any:
        if not self.params_json_schema:
            _schema = function_to_json(self._run, name=self.name, description=self.description)
            self.params_json_schema = _schema["function"]["parameters"]
        return self

    def to_openai(self) -> dict:
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description or "",
                "parameters": self.params_json_schema,
            },
        }

    def to_openai_new(self) -> dict:
        return {
            "type": "function",
            "name": self.name,
            "description": self.description or "",
            "parameters": self.params_json_schema,
        }

    def _cast_response_to_str(self, response) -> str:
        if isinstance(response, str):
            return response
        elif isinstance(response, BaseModel):
            return response.model_dump_json(exclude_none=True)
        return json.dumps(response, ensure_ascii=False)

    @abstractmethod
    def _run(self, **kwargs: Any) -> Any:
        """Actual implementation of the tool."""

    def run(self, **kwargs: Any) -> str:
        response = self._run(**kwargs)
        return self._cast_response_to_str(response)

    async def arun(self, **kwargs: Any) -> str:
        response = asyncio.run(self._run(**kwargs))
        return self._cast_response_to_str(response)



from pydantic import BaseModel, Field


class UsageDetails(BaseModel):
    cached_tokens: int = 0
    reasoning_tokens: int = 0

    def add(self, other: "UsageDetails") -> None:
        self.cached_tokens += other.cached_tokens
        self.reasoning_tokens += other.reasoning_tokens

class Usage(BaseModel):
    requests: int = 0
    input_tokens: int = 0
    output_tokens: int = 0
    total_tokens: int = 0

    input_tokens_details: UsageDetails = Field(default_factory=UsageDetails)
    output_tokens_details: UsageDetails = Field(default_factory=UsageDetails)

    def add(self, other: "Usage") -> None:
        self.requests += other.requests if other.requests else 0
        self.input_tokens += other.input_tokens if other.input_tokens else 0
        self.output_tokens += other.output_tokens if other.output_tokens else 0
        self.total_tokens += other.total_tokens if other.total_tokens else 0

        if other.input_tokens_details:
            self.input_tokens_details.add(other.input_tokens_details)
        if other.output_tokens_details:
            self.output_tokens_details.add(other.output_tokens_details)



import json
import os
from typing import Any, Dict, List

import requests
from pydantic import Field

from gwenflow.tools import BaseTool


class WebSearchTool(BaseTool):
    name: str = "WebSearchTool"
    description: str = "Searches the web for information related to a given query."

    search_engine_id: str = Field(default_factory=lambda: os.getenv("WEBSEARCH_SEARCH_ENGINE_ID"))
    api_key: str = Field(default_factory=lambda: os.getenv("WEBSEARCH_API_KEY"))
    base_url: str = "https://www.googleapis.com/customsearch/v1"

    def _run(self, query: str = Field(description="The search query."), num_results: int = 10) -> Dict[str, Any]:
        """Effectue une recherche web en utilisant l'API Google Custom Search.

        Args:
            query: La requête de recherche
            num_results: Le nombre de résultats à retourner (max 10)

        Returns:
            Dict contenant les résultats de recherche
        """
        try:
            params = {
                "key": self.api_key,
                "cx": self.search_engine_id,
                "q": query,
                "num": min(num_results, 10),
                "safe": "active",
            }

            response = requests.get(self.base_url, params=params)
            response.raise_for_status()

            data = response.json()

            results = self._parse_results(data)

            return {
                "success": True,
                "query": query,
                "total_results": data.get("searchInformation", {}).get("totalResults", "0"),
                "search_time": data.get("searchInformation", {}).get("searchTime", "0"),
                "results": results,
            }

        except requests.exceptions.RequestException as e:
            return {"success": False, "error": f"Erreur de requête: {str(e)}", "query": query, "results": []}
        except json.JSONDecodeError as e:
            return {"success": False, "error": f"Erreur de parsing JSON: {str(e)}", "query": query, "results": []}
        except Exception as e:
            return {"success": False, "error": f"Erreur inattendue: {str(e)}", "query": query, "results": []}

    def _parse_results(self, data: Dict[str, Any]) -> List[Dict[str, str]]:
        """Parse les résultats de l'API Google Custom Search.

        Args:
            data: Réponse JSON de l'API

        Returns:
            Liste des résultats parsés
        """
        results = []
        items = data.get("items", [])

        for item in items:
            result = {
                "title": item.get("title", ""),
                "link": item.get("link", ""),
                "snippet": item.get("snippet", ""),
                "display_link": item.get("displayLink", ""),
                "formatted_url": item.get("formattedUrl", ""),
            }

            if "pagemap" in item:
                pagemap = item["pagemap"]
                if "metatags" in pagemap and pagemap["metatags"]:
                    metatag = pagemap["metatags"][0]
                    result["description"] = metatag.get("og:description", result["snippet"])
                    result["image"] = metatag.get("og:image", "")

            results.append(result)

        return results


if __name__ == "__main__":
    search_tool = WebSearchTool()

    results = search_tool._run("Python programmation tutoriel")

    if results["success"]:
        print(f"Requête: {results['query']}")
        print(f"Nombre total de résultats: {results['total_results']}")
        print(f"Temps de recherche: {results['search_time']} secondes")
        print("\nRésultats:")

        for i, result in enumerate(results["results"], 1):
            print(f"\n{i}. {result['title']}")
            print(f"   URL: {result['link']}")
            print(f"   Description: {result['snippet'][:100]}...")
    else:
        print(f"Erreur: {results['error']}")



from gwenflow.readers.directory import SimpleDirectoryReader
from gwenflow.readers.docx import DocxReader
from gwenflow.readers.json import JSONReader
from gwenflow.readers.pdf import PDFReader
from gwenflow.readers.text import TextReader
from gwenflow.readers.website import WebsiteReader

__all__ = ["SimpleDirectoryReader", "TextReader", "JSONReader", "PDFReader", "WebsiteReader", "DocxReader"]



from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator

from gwenflow.agents import Agent
from gwenflow.llms import ChatBase, ChatOpenAI
from gwenflow.logger import logger
from gwenflow.types import Message

PROMPT_REFORMULATE = """\
# Instructions
- You are an AI assistant reading a current user query and chat_history.
- Given the chat_history, and current user's query, infer the user's intent expressed in the current user query.
- Once you infer the intent, respond with a task that can be used to with an Agent for the current user's query based on the intent
- Be specific in what the user is asking about, but disregard parts of the chat history that are not relevant to the user's intent.
- Provide responses in JSON format

# Examples
Example 1:
With a conversation like below:
```
 - user: are the trailwalker shoes waterproof?
 - assistant: Yes, the TrailWalker Hiking Shoes are waterproof. They are designed with a durable and waterproof construction to withstand various terrains and weather conditions.
 - user: how much do they cost?
```
Respond with:
{
    "intent": "The user wants to know how much the Trailwalker Hiking Shoes cost.",
    "task": "price of Trailwalker Hiking Shoes"
}

Example 2:
With a conversation like below:
```
 - user: are the trailwalker shoes waterproof?
 - assistant: Yes, the TrailWalker Hiking Shoes are waterproof. They are designed with a durable and waterproof construction to withstand various terrains and weather conditions.
 - user: how much do they cost?
 - assistant: The TrailWalker Hiking Shoes are priced at $110.
 - user: do you have waterproof tents?
 - assistant: Yes, we have waterproof tents available. Can you please provide more information about the type or size of tent you are looking for?
 - user: which is your most waterproof tent?
 - assistant: Our most waterproof tent is the Alpine Explorer Tent. It is designed with a waterproof material and has a rainfly with a waterproof rating of 3000mm. This tent provides reliable protection against rain and moisture.
 - user: how much does it cost?
```
Respond with:
{
    "intent": "The user would like to know how much the Alpine Explorer Tent costs.",
    "task": "price of Alpine Explorer Tent"
}

Return the task for the messages in the following conversation:

"""


class ChatAgent(BaseModel):
    llm: Optional[ChatBase] = Field(None, validate_default=True)
    agent: Agent

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    @field_validator("llm", mode="before")
    @classmethod
    def set_llm(cls, v: Optional[Any]) -> Any:
        llm = v or ChatOpenAI(model="gpt-4o-mini")
        return llm

    def run(self, messages: Union[str, List[Message], List[Dict[str, str]]]) -> Any:
        messages = self.llm._cast_messages(messages)

        conversation = []
        for message in messages:
            conversation.append(f"- {message.role}: {message.content}")
        conversation = "\n".join(conversation)
        prompt = PROMPT_REFORMULATE + f"<conversation>\n{conversation}\n</conversation>"

        self.llm.response_format = {"type": "json_object"}
        response = self.llm.invoke(prompt)
        response = response.choices[0].message.content

        logger.debug("Task: " + response["task"])

        return self.agent.run(response["task"])



import hashlib
import io
import os
import pickle
from typing import Optional

import numpy as np

from gwenflow.utils.aws import (
    aws_s3_download_in_buffer,
    aws_s3_is_file,
    aws_s3_upload_fileobj,
    aws_s3_uri_to_bucket_key,
)

try:
    import faiss
except ImportError as exc:
    raise ImportError("`faiss` is not installed.") from exc

from gwenflow.embeddings import Embeddings, GwenlakeEmbeddings
from gwenflow.logger import logger
from gwenflow.reranker import Reranker
from gwenflow.types import Document
from gwenflow.vector_stores.base import VectorStoreBase


class FAISS(VectorStoreBase):
    def __init__(
        self,
        filename: str,
        embeddings: Optional[Embeddings] = None,
        reranker: Optional[Reranker] = None,
    ):
        # Embedder
        self.embeddings = embeddings or GwenlakeEmbeddings()

        # reranker
        self.reranker = reranker

        # name
        self.filename = filename
        if not self.filename.endswith(".pkl"):
            self.filename = self.filename + ".pkl"

        self.index = None
        self.metadata = []
        self.create()

    def create(self):
        if not self.exists():
            self.index = faiss.IndexFlatL2(self.embeddings.dimensions)
            self.metadata = []
            self.save()
        else:
            self.load()

    def exists(self) -> bool:
        if self.filename.startswith("s3://"):
            bucket, key = aws_s3_uri_to_bucket_key(self.filename)
            return aws_s3_is_file(bucket=bucket, key=key)
        else:
            return os.path.isfile(self.filename)

    def get_collections(self) -> list:
        return []

    def insert(self, documents: list[Document]):
        logger.info(f"Embedding {len(documents)} documents")
        embeddings = self.embeddings.embed_documents([document.content for document in documents])
        embeddings = np.array(embeddings, dtype="float32")

        logger.info(f"Inserting {len(documents)} documents into index")
        data = []
        for document in documents:
            if document.id is None:
                document.id = hashlib.md5(document.content.encode(), usedforsecurity=False).hexdigest()
            data.append(document.model_dump())

        if len(documents) > 0:
            self.index.add(embeddings)
            self.metadata.extend(data)
            self.save()

    def search(self, query: str, limit: int = 5) -> list[Document]:
        if not self.index:
            logger.error("Error no index.")
            return []

        query_embedding = self.embeddings.embed_query(query)
        if query_embedding is None:
            logger.error(f"Error getting embedding for Query: {query}")
            return []
        query_embedding = np.array([query_embedding], dtype="float32")

        D, I = self.index.search(query_embedding, k=limit)  # noqa: E741, N806

        documents = []
        for idx, score in zip(I[0], D[0], strict=False):
            if idx == -1:
                continue
            document = self.metadata[idx].copy()
            document.pop("embedding")
            document = Document(**document)
            document.score = score
            documents.append(document)

        if self.reranker:
            documents = self.reranker.rerank(query=query, documents=documents)

        return documents

    def save(self):
        try:
            faiss_data = {"index": self.index, "metadata": self.metadata}
            if self.filename.startswith("s3://"):
                logger.info(f"Saving FAISS index to S3 at {self.filename} ...")
                bucket, key = aws_s3_uri_to_bucket_key(self.filename)
                buffer = io.BytesIO()
                pickle.dump(faiss_data, buffer)
                buffer.seek(0)
                aws_s3_upload_fileobj(bucket=bucket, key=key, fileobj=buffer)
            else:
                logger.info(f"Saving FAISS index locally at {self.filename} ...")
                with open(self.filename, "wb") as f:
                    pickle.dump(faiss_data, f)
            return True
        except Exception as e:
            logger.error(e)
        return False

    def load(self):
        try:
            logger.info(f"Loading FAISS index from: {self.filename} ...")
            if self.filename.startswith("s3://"):
                bucket, key = aws_s3_uri_to_bucket_key(self.filename)
                buffer = aws_s3_download_in_buffer(bucket=bucket, key=key)
                faiss_data = pickle.load(buffer)
            else:
                with open(self.filename, "rb") as f:
                    faiss_data = pickle.load(f)

            self.index = faiss_data["index"]
            self.metadata = faiss_data["metadata"]
        except Exception as e:
            logger.error(e)

    def drop(self):
        try:
            os.remove(self.filename)
            self.index = faiss.IndexFlatL2(self.embeddings.dimensions)
            self.metadata = []
            return True
        except Exception as e:
            logger.error(e)
        return False

    def count(self) -> int:
        return 0

    def info(self) -> dict:
        return {}

    def delete(self, id: int):
        return False

    def delete_files(self, ids: list[str]) -> bool:
        logger.info(f"Deleting file with document_id: {ids}")

        keep_indices = [
            i for i, doc in enumerate(self.metadata) if doc.get("metadata", {}).get("document_id") not in ids
        ]

        if len(keep_indices) == len(self.metadata):
            logger.warning(f"No chunks found for document_id: {ids}")
            return False

        new_index = faiss.IndexFlatL2(self.embeddings.dimensions)
        new_metadata = []

        for i in keep_indices:
            vector = self.index.reconstruct(i)
            new_index.add(np.array([vector], dtype="float32"))
            new_metadata.append(self.metadata[i])

        self.index = new_index
        self.metadata = new_metadata
        self.save()

        return True

    def get(self, id: int) -> dict:
        return None

    def list(self, filters: dict = None, limit: int = 100) -> list:
        return []



from typing import Annotated, Any, List, Literal, Optional, Union

from pydantic import BaseModel, Field, RootModel

from gwenflow.tools import BaseTool
from gwenflow.types import Usage


class ResponseContent(BaseModel):
    type: Literal['output_text', 'summary_text']
    text: Any
    annotations: Optional[List] = Field(default_factory=list)


class ResponseOutputItem(BaseModel):
    id: str
    type: Literal["reasoning", "message"]
    status: Optional[str] = None
    summary: Optional[List[ResponseContent]] = None
    content: Optional[List[ResponseContent]] = None
    role: Optional[str] = None


class ReasoningItem(BaseModel):
    effort: Literal['low', 'medium', 'high']
    summary: Literal['auto', 'concise', 'detailed']


class Response(BaseModel):
    model_config = {"extra": "ignore"}
    id: str
    object: Literal["response"]
    model: str
    status: str
    created_at: float
    completed_at: Optional[float] = None
    output: List[ResponseOutputItem]
    usage: Usage
    reasoning: Optional[ReasoningItem] = None
    text_format: Optional[Any] = None
    tools: Optional[List[BaseTool]] = None
    tool_choice: Optional[Literal['none', 'auto', 'required']] = None
    temperature: Optional[float] = None
    top_p : Optional[float] = None

    def get_text(self) -> str:
        for item in self.output:
            if item.type == "message" and item.content:
                return item.content[0].text
        return ""

    def get_reasoning(self) -> str:
        for item in self.output:
            if item.type == "reasoning" and item.summary:
                return item.summary[0].text
        return ""


class Logprob(BaseModel):
    token: str
    logprob: float
    bytes: Optional[List[int]] = None

class ResponseDeltaEventBase(BaseModel):
    item_id: str
    sequence_number: int


class ResponseReasoningDeltaEvent(ResponseDeltaEventBase):
    type: Literal["response.reasoning_summary_text.delta"]
    delta: str

class ResponseTextDeltaEvent(ResponseDeltaEventBase):
    type: Literal["response.output_text.delta"]
    delta: str
    logprobs: Optional[List[Logprob]] = Field(default_factory=list)


class ResponseDoneEvent(BaseModel):
    type: Literal["response.completed"]
    response: Response

AnyResponseEvent = Annotated[
    Union[
        ResponseReasoningDeltaEvent,
        ResponseTextDeltaEvent,
        ResponseDoneEvent
    ],
    Field(discriminator="type")
]

class ResponseEvent(RootModel):
    root: AnyResponseEvent



import enum
import hashlib
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, model_validator


class DocumentCreationMode(str, enum.Enum):
    """Enumeration for document creation modes."""

    ONE_DOC_PER_FILE = "one-doc-per-file"
    ONE_DOC_PER_PAGE = "one-doc-per-page"
    ONE_DOC_PER_ELEMENT = "one-doc-per-element"


class Document(BaseModel):
    """Base class for Documents."""

    id: Optional[str] = None
    content: str
    metadata: Dict[str, Any] = {}
    embedding: List[float] = None
    score: Optional[float] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="after")
    def set_id(self) -> "Document":
        if self.id is None:
            self.id = hashlib.md5(self.content.encode(), usedforsecurity=False).hexdigest()
        return self

    def to_dict(self) -> Dict[str, Any]:
        """Returns a dictionary representation of the document."""
        return self.model_dump(include={"id", "content", "metadata", "score"}, exclude_none=True)

    @classmethod
    def from_dict(cls, document: Dict[str, Any]) -> "Document":
        """Returns a Document object from a dictionary representation."""
        return cls.model_validate(**document)

    @classmethod
    def from_json(cls, document: str) -> "Document":
        """Returns a Document object from a json string representation."""
        return cls.model_validate_json(document)



from gwenflow.agents import Agent, ChatAgent, ReactAgent
from gwenflow.exceptions import (
    GwenflowException,
    MaxTurnsExceeded,
    ModelBehaviorError,
    UserError,
)
from gwenflow.flows import AutoFlow, Flow
from gwenflow.llms import ChatAzureOpenAI, ChatGwenlake, ChatOllama, ChatOpenAI
from gwenflow.logger import logger, set_log_level_to_debug
from gwenflow.readers import SimpleDirectoryReader
from gwenflow.retriever import Retriever
from gwenflow.tools import BaseTool, FunctionTool
from gwenflow.types import Document, Message
from gwenflow.telemetry import TelemetryBase

__all__ = [
    "logger",
    "set_log_level_to_debug",
    "GwenflowException",
    "MaxTurnsExceeded",
    "ModelBehaviorError",
    "UserError",
    "ChatGwenlake",
    "ChatOpenAI",
    "ChatAzureOpenAI",
    "ChatOllama",
    "Document",
    "Message",
    "SimpleDirectoryReader",
    "Retriever",
    "Agent",
    "ReactAgent",
    "ChatAgent",
    "BaseTool",
    "FunctionTool",
    "Flow",
    "AutoFlow",
    "TelemetryBase",
]



import os
from typing import Optional

from httpx import AsyncClient, Client
from pydantic import BaseModel

from gwenflow.version import __version__


class Api(BaseModel):
    base_url: Optional[str] = "https://api.gwenlake.com"
    api_key: Optional[str] = None
    timeout: Optional[int] = 300

    def get_headers(self) -> dict:
        api_key = self.api_key
        if api_key is None:
            api_key = os.environ.get("GWENLAKE_API_KEY")
        if api_key is None:
            raise ValueError(
                "The api_key client option must be set either by passing api_key to the client or by setting the GWENLAKE_API_KEY environment variable"
            )
        headers = {
            "Content-Type": "application/json",
            "user-agent": f"gwenflow/{__version__}",
            "Authorization": f"Bearer {api_key}",
        }
        return headers

    @property
    def client(self) -> Client:
        return Client(
            base_url=self.base_url,
            headers=self.get_headers(),
            timeout=self.timeout,
        )

    @property
    def async_client(self) -> AsyncClient:
        return AsyncClient(
            base_url=self.base_url,
            headers=self.get_headers(),
            timeout=self.timeout,
        )


api = Api()



import json
import os
from collections.abc import AsyncIterator
from typing import Any, Dict, Iterator, List, Optional, Union

from openai import AsyncOpenAI, OpenAI
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types.chat.chat_completion_chunk import ChoiceDelta

from gwenflow.llms.base import ChatBase
from gwenflow.logger import logger
from gwenflow.telemetry import tracer
from gwenflow.types import ItemHelpers, Message, ModelResponse, ToolCall, Usage
from gwenflow.utils import extract_json_str


class ChatOpenAI(ChatBase):

    model: str = "gpt-4o-mini"

    # model parameters
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    n: Optional[int] = None
    stop: Optional[Union[str, List[str]]] = None
    max_completion_tokens: Optional[int] = None
    max_tokens: Optional[int] = None
    presence_penalty: Optional[float] = None
    frequency_penalty: Optional[float] = None
    logit_bias: Optional[Dict[int, float]] = None
    response_format: Optional[Dict[str, Any]] = None
    seed: Optional[int] = None
    logprobs: Optional[bool] = None
    top_logprobs: Optional[int] = None

    # clients
    client: Optional[OpenAI] = None
    async_client: Optional[AsyncOpenAI] = None

    # client parameters
    api_key: Optional[str] = None
    organization: Optional[str] = None
    base_url: Optional[str] = None
    timeout: Optional[Union[float, int]] = None
    max_retries: Optional[int] = None

    def _get_client_params(self) -> Dict[str, Any]:

        api_key = self.api_key
        if api_key is None:
            api_key = os.environ.get("OPENAI_API_KEY")
        if api_key is None:
            raise ValueError(
                "The api_key client option must be set either by passing api_key to the client or by setting the OPENAI_API_KEY environment variable"
            )

        organization = self.organization
        if organization is None:
            organization = os.environ.get('OPENAI_ORG_ID')

        client_params = {
            "api_key": api_key,
            "organization": organization,
            "base_url": self.base_url,
            "timeout": self.timeout,
            "max_retries": self.max_retries,
        }

        client_params = {k: v for k, v in client_params.items() if v is not None}

        return client_params

    @property
    def _model_params(self) -> Dict[str, Any]:

        model_params = {
            "temperature": self.temperature,
            "top_p": self.top_p,
            "n": self.n,
            "stop": self.stop,
            "max_tokens": self.max_tokens or self.max_completion_tokens,
            "presence_penalty": self.presence_penalty,
            "frequency_penalty": self.frequency_penalty,
            "logit_bias": self.logit_bias,
            "response_format": self.response_format,
            "seed": self.seed,
            "logprobs": self.logprobs,
            "top_logprobs": self.top_logprobs,
        }

        if self.tools and self.tool_type == "fncall":
            model_params["tools"] = [tool.to_openai() for tool in self.tools]
            model_params["tool_choice"] = self.tool_choice or "auto"

        model_params = {k: v for k, v in model_params.items() if v is not None}

        return model_params

    def get_client(self) -> OpenAI:
        if self.client:
            return self.client
        client_params = self._get_client_params()
        self.client = OpenAI(**client_params)
        return self.client

    def get_async_client(self) -> AsyncOpenAI:
        if self.async_client:
            return self.async_client
        client_params = self._get_client_params()
        self.async_client = AsyncOpenAI(**client_params)
        return self.async_client

    def _format_response(self, response: str, response_format: dict = None) -> Any:
        """Process the response."""

        if response_format.get("type") == "json_object":
            try:
                json_str = extract_json_str(response)
                text_response = json.loads(json_str)
                return text_response
            except:
                pass

        return response

    def _format_message(self, message: Message) -> Dict[str, Any]:
        """Format a message into the format expected by OpenAI."""
        return message.to_openai()

    def _get_openai_usage(self, completion: Union[ChatCompletion, ChatCompletionChunk]) -> Usage:
        if not completion.usage:
            return None
        return Usage(
            requests=1,
            input_tokens=completion.usage.prompt_tokens,
            output_tokens=completion.usage.completion_tokens,
            total_tokens=completion.usage.total_tokens,
        )

    def _parse_response(self, completion: ChatCompletion) -> ModelResponse:
        model_response = ModelResponse(
            role=completion.choices[0].message.role,
            content=completion.choices[0].message.content,
            finish_reason="stop",
            usage=self._get_openai_usage(completion),
        )

        if hasattr(completion.choices[0].message, 'reasoning_content'):
            model_response.reasoning_content = completion.choices[0].message.reasoning_content

        tool_calls = completion.choices[0].message.tool_calls
        if tool_calls is not None and len(tool_calls) > 0:
            try:
                model_response.tool_calls = [ToolCall(**t.model_dump()) for t in tool_calls]
            except Exception as e:
                logger.warning(f"Error processing tool calls: {e}")

        if self.response_format:
            model_response.parsed = self._format_response(completion.choices[0].message.content, response_format=self.response_format)

        return model_response

    @tracer.llm(name="LLM Invoke")
    def invoke(self, input: Union[str, List[Message], List[Dict[str, str]]]) -> ModelResponse:
        try:
            messages_for_model = ItemHelpers.input_to_message_list(input)
            response: ChatCompletion = self.get_client().chat.completions.create(
                model=self.model,
                messages=[self._format_message(m) for m in messages_for_model],
                **self._model_params,
            )
        except Exception as e:
            raise RuntimeError(f"Error in calling openai API: {e}")

        return self._parse_response(response)

    @tracer.llm(name="LLM Async Invoke")
    async def ainvoke(self, input: Union[str, List[Message], List[Dict[str, str]]]) -> ModelResponse:
        try:
            messages_for_model = ItemHelpers.input_to_message_list(input)
            response: ChatCompletion = await self.get_async_client().chat.completions.create(
                model=self.model,
                messages=[self._format_message(m) for m in messages_for_model],
                **self._model_params,
            )
        except Exception as e:
            raise RuntimeError(f"Error in calling openai API: {e}")

        return self._parse_response(response)

    @tracer.llm(name="LLM Stream")
    def stream(self, input: Union[str, List[Message], List[Dict[str, str]]]) -> Iterator[ModelResponse]:
        try:
            messages_for_model = ItemHelpers.input_to_message_list(input)
            completion = self.get_client().chat.completions.create(
                model=self.model,
                messages=[self._format_message(m) for m in messages_for_model],
                stream=True,
                stream_options={"include_usage": True},
                **self._model_params,
            )

            _full_tool_calls = []

            for chunk in completion:

                response = ModelResponse(role="assistant")

                if chunk.choices:

                    delta: ChoiceDelta = chunk.choices[0].delta

                    if hasattr(delta, 'content') and delta.content:
                        response.content = delta.content

                    if hasattr(delta, 'tool_calls') and delta.tool_calls:
                        for tool_call in delta.tool_calls or []:
                            if _full_tool_calls and (not tool_call.id or tool_call.id == _full_tool_calls[-1].id):
                                if tool_call.function.name:
                                    _full_tool_calls[-1].function.name += tool_call.function.name
                                if tool_call.function.arguments:
                                    _full_tool_calls[-1].function.arguments += tool_call.function.arguments
                            else:
                                _full_tool_calls.append(
                                    ToolCall(**tool_call.model_dump())
                                )

                    if _full_tool_calls:
                        response.tool_calls = _full_tool_calls

                    if response.content or response.reasoning_content:
                        yield response
                    elif len(response.tool_calls)>0 and chunk.choices[0].finish_reason == "tool_calls":
                        yield response

                else:
                    if hasattr(response, 'usage'):
                        response.usage = self._get_openai_usage(chunk)
                        yield response

        except Exception as e:
            raise RuntimeError(f"Error in calling openai API: {e}")

    @tracer.llm(name="LLM Async Astream")
    async def astream(self, input: Union[str, List[Message], List[Dict[str, str]]]) -> AsyncIterator[ModelResponse]:
        try:
            messages_for_model = ItemHelpers.input_to_message_list(input)
            completion = await self.get_async_client().chat.completions.create(
                model=self.model,
                messages=[self._format_message(m) for m in messages_for_model],
                stream=True,
                stream_options={"include_usage": True},
                **self._model_params,
            )

            _full_tool_calls = []

            async for chunk in completion:

                response = ModelResponse(role="assistant")

                if chunk.choices:

                    delta: ChoiceDelta = chunk.choices[0].delta

                    if hasattr(delta, 'content') and delta.content:
                        response.content = delta.content

                    if hasattr(delta, 'tool_calls') and delta.tool_calls:
                        for tool_call in delta.tool_calls or []:
                            if _full_tool_calls and (not tool_call.id or tool_call.id == _full_tool_calls[-1].id):
                                if tool_call.function.name:
                                    _full_tool_calls[-1].function.name += tool_call.function.name
                                if tool_call.function.arguments:
                                    _full_tool_calls[-1].function.arguments += tool_call.function.arguments
                            else:
                                _full_tool_calls.append(
                                    ToolCall(**tool_call.model_dump())
                                )

                    if _full_tool_calls:
                        response.tool_calls = _full_tool_calls

                    if response.content or response.reasoning_content:
                        yield response
                    elif len(response.tool_calls)>0 and chunk.choices[0].finish_reason == "tool_calls":
                        yield response

                else:
                    if hasattr(response, 'usage'):
                        response.usage = self._get_openai_usage(chunk)
                        yield response

        except Exception as e:
            raise RuntimeError(f"Error in calling openai API: {e}")


from typing import Optional

from pydantic import Field, field_validator

from gwenflow.logger import logger
from gwenflow.memory.base import BaseChatMemory
from gwenflow.types import Message
from gwenflow.utils.tokens import keep_tokens_from_text

DEFAULT_TOKEN_LIMIT = 8192
DEFAULT_TOKEN_LIMIT_RATIO = 0.75
MAX_MESSAGE_CONTENT = 0.5


class ChatMemoryBuffer(BaseChatMemory):
    token_limit: Optional[int] = Field(None, validate_default=True)

    @field_validator("token_limit", mode="before")
    @classmethod
    def set_token_limit(cls, v: Optional[int]) -> int:
        token_limit = v or int(DEFAULT_TOKEN_LIMIT * DEFAULT_TOKEN_LIMIT_RATIO)
        return token_limit

    def get(self):
        initial_token_count = 0

        if self.system_prompt:
            initial_token_count = self._token_count_for_messages([Message(role="system", content=self.system_prompt)])
        if initial_token_count > self.token_limit:
            raise ValueError("Initial token count exceeds token limit")

        chat_history = self.messages
        message_count = len(chat_history)

        cur_messages = chat_history[-message_count:]
        token_count = self._token_count_for_messages(cur_messages) + initial_token_count

        # pre filter very large messages
        for index, message in enumerate(chat_history):
            if self._token_count_for_messages([message]) > (MAX_MESSAGE_CONTENT * self.token_limit):
                chat_history[index].content = keep_tokens_from_text(
                    message.content,
                    token_limit=int(MAX_MESSAGE_CONTENT * self.token_limit),
                    tokenizer_fn=self.tokenizer_fn,
                )

        # keep messages
        while token_count > self.token_limit and message_count > 1:
            message_count -= 1
            while chat_history[-message_count].role in ("tool", "assistant"):
                message_count -= 1
            cur_messages = chat_history[-message_count:]
            token_count = self._token_count_for_messages(cur_messages) + initial_token_count

        if token_count > self.token_limit:
            logger.warning("Token limit exceeded.")
            if self.system_prompt:
                return [Message(role="system", content=self.system_prompt)]
            return []

        if self.system_prompt:
            return [Message(role="system", content=self.system_prompt)] + chat_history[-message_count:]

        return chat_history[-message_count:]
